import { useRef, useReducer, useEffect, useLayoutEffect } from 'react';

const forceUpdateReducer = state => state + 1; // For server-side rendering: https://github.com/react-spring/zustand/pull/34


const useIsoLayoutEffect = typeof window === 'undefined' ? useEffect : useLayoutEffect;
function create(createState) {
  // All listeners are wrapped in a function with the signature: () => void
  const listeners = [];
  let state;
  let renderCount = 0; // Returns an int for a component based on render order.

  function useRenderId() {
    const renderIdRef = useRef();

    if (!renderIdRef.current) {
      renderIdRef.current = renderCount++;
    }

    return renderIdRef.current;
  }

  const setState = partial => {
    const partialState = typeof partial === 'function' ? partial(state) : partial;

    if (partialState !== state) {
      state = Object.assign({}, state, partialState);
      listeners.forEach(listener => listener());
    }
  };

  const getState = () => state;

  const subscribe = (listener, options = {}) => {
    if (!('currentSlice' in options)) {
      options.currentSlice = (options.selector || getState)(state);
    } // subscribe can be called externally without passing in a listenerIndex so
    // we need to assign it a default index.


    const {
      listenerIndex = renderCount++
    } = options;

    const listenerWrapper = () => {
      // Access the current values of the options object in listenerWrapper.
      // We rely on this because options is mutated in useStore.
      const {
        selector = getState,
        equalityFn = Object.is
      } = options; // Selector or equality function could throw but we don't want to stop
      // the listener from being called.
      // https://github.com/react-spring/zustand/pull/37

      try {
        const newStateSlice = selector(state);

        if (!equalityFn(options.currentSlice, newStateSlice)) {
          listener(options.currentSlice = newStateSlice);
        }
      } catch (error) {
        options.subscribeError = error;
        listener(null, error);
      }
    };

    listeners[listenerIndex] = listenerWrapper; // Intentially using delete because shortening the length of the listeners
    // array would result in listenerIndex not accessing the correct listener.
    // This means listeners should be considered a sparce array.

    return () => delete listeners[listenerIndex];
  };

  const apiSubscribe = (listener, selector, equalityFn) => subscribe(listener, {
    selector,
    equalityFn
  });

  const destroy = () => listeners.length = 0;

  const useStore = (selector = getState, equalityFn = Object.is) => {
    const listenerIndex = useRenderId();
    const optionsRef = useRef();

    if (!optionsRef.current) {
      optionsRef.current = {
        selector,
        equalityFn,
        currentSlice: selector(state),
        listenerIndex
      };
    }

    const options = optionsRef.current; // Update state slice if selector has changed or subscriber errored.

    if (selector !== options.selector || options.subscribeError) {
      const newStateSlice = selector(state);

      if (!equalityFn(options.currentSlice, newStateSlice)) {
        options.currentSlice = newStateSlice;
      }
    }

    useIsoLayoutEffect(() => {
      options.selector = selector;
      options.equalityFn = equalityFn;
      options.subscribeError = undefined;
    });
    const forceUpdate = useReducer(forceUpdateReducer, 0)[1];
    useIsoLayoutEffect(() => subscribe(forceUpdate, options), []);
    return options.currentSlice;
  };

  const api = {
    setState,
    getState,
    subscribe: apiSubscribe,
    destroy
  };
  state = createState(setState, getState, api);
  return [useStore, api];
}

export default create;
export { create };
